from django.contrib import admin
from django import forms

from . import global_preferences_registry
from .models import GlobalPreferenceModel, UserPreferenceModel
from .forms import GlobalSinglePreferenceForm, UserSinglePreferenceForm, SinglePerInstancePreferenceForm


class DynamicPreferenceAdmin(admin.ModelAdmin):
    list_display = ('name', 'section', 'raw_value')
    list_editable = ('raw_value',)
    search_fields = ['name', 'section', 'raw_value']
    list_filter = ('section',)

    def get_changelist_form(self, request, **kwargs):
        return self.changelist_form


class GlobalPreferenceAdmin(DynamicPreferenceAdmin):
    form = GlobalSinglePreferenceForm
    changelist_form = GlobalSinglePreferenceForm

    def get_queryset(self, *args, **kwargs):
        # Instanciate default prefs
        manager = global_preferences_registry.manager()
        manager.all()
        return super(GlobalPreferenceAdmin, self).get_queryset(*args, **kwargs)
        
admin.site.register(GlobalPreferenceModel, GlobalPreferenceAdmin)


class PerInstancePreferenceAdmin(DynamicPreferenceAdmin):
    list_display = ('instance',) + DynamicPreferenceAdmin.list_display
    raw_id_fields = ('instance',)
    form = SinglePerInstancePreferenceForm

class UserPreferenceAdmin(PerInstancePreferenceAdmin):
    search_fields = ['instance__username'] + DynamicPreferenceAdmin.search_fields
    form = UserSinglePreferenceForm
    changelist_form = UserSinglePreferenceForm

admin.site.register(UserPreferenceModel, UserPreferenceAdmin)



from . import user_preferences_registry as upr, global_preferences_registry as gpr

def global_preferences(request):
    """
        Pass the values of global preferences to template context.
        You can then access value with `global_preferences.<section>.<name>`
    """
    manager = gpr.manager()
    return {'global_preferences': manager.all()}



from django.core.management.base import BaseCommand, CommandError
from dynamic_preferences.models import GlobalPreferenceModel, UserPreferenceModel
from dynamic_preferences import global_preferences_registry
from dynamic_preferences.registries import preference_models


def delete_preferences(queryset):
    """
    Delete preferences objects if they are not present in registry. Return a list of deleted objects
    """
    deleted = []

    # Iterate through preferences. If an error is raised when accessing preference object, just delete it
    for p in queryset:
        try:
            pref = p.preference
        except KeyError:
            p.delete()
            deleted.append(p)

    return deleted


class Command(BaseCommand):
    help = "Find and delete preferences from database if they don't exist in registries. Create preferences that are " \
           "not present in database"

    def handle(self, *args, **options):

        # Create needed preferences
        # Global
        print('Creating missing global preferences...')
        manager = global_preferences_registry.manager()
        manager.all()

        deleted = delete_preferences(GlobalPreferenceModel.objects.all())
        print("Deleted {0} global preferences".format(len(deleted)))

        for preference_model, registry in preference_models.items():
            deleted = delete_preferences(preference_model.objects.all())
            print("Deleted {0} {1} preferences".format(len(deleted), preference_model.__class__.__name__))
            if not hasattr(preference_model, 'get_instance_model'):
                continue

            print('Creating missing preferences for {0} model...'.format(preference_model.get_instance_model().__name__))
            for instance in preference_model.get_instance_model().objects.all():
                instance.preferences.all()



import collections

from .settings import preferences_settings
from .exceptions import CachedValueNotFound, DoesNotExist


class PreferencesManager(collections.Mapping):

    """Handle retrieving / caching of preferences"""

    def __init__(self, model, registry, **kwargs):

        self.model = model

        self.registry = registry
        self.queryset = self.model.objects.all()
        self.instance = kwargs.get('instance')
        if self.instance:
            self.queryset = self.queryset.filter(instance=self.instance)

    @property
    def cache(self):
        from django.core.cache import caches
        return caches['default']

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        section, name = self.parse_lookup(key)
        self.update_db_pref(section=section, name=name, value=value)

    def __repr__(self):
        return repr(self.all())

    def __iter__(self):
        return self.all().__iter__()

    def __len__(self):
        return len(self.all())

    def get_cache_key(self, section, name):
        """Return the cache key corresponding to a given preference"""
        if not self.instance:
            return 'dynamic_preferences_{0}_{1}_{2}'.format(self.model.__name__, section, name)
        return 'dynamic_preferences_{0}_{1}_{2}_{3}'.format(self.model.__name__, section, name, self.instance.pk)

    def from_cache(self, section, name):
        """Return a preference raw_value from cache"""
        cached_value = self.cache.get(
            self.get_cache_key(section, name), CachedValueNotFound)

        if cached_value is CachedValueNotFound:
            raise CachedValueNotFound
        return self.registry.get(section=section, name=name).serializer.deserialize(cached_value)

    def to_cache(self, pref):
        """Update/create the cache value for the given preference model instance"""
        self.cache.set(
            self.get_cache_key(pref.section, pref.name), pref.raw_value, None)

    def pref_obj(self, section, name):
        return self.registry.get(section=section, name=name)

    def parse_lookup(self, lookup):
        try:
            section, name = lookup.split(
                preferences_settings.SECTION_KEY_SEPARATOR)
        except ValueError:
            name = lookup
            section = None
        return section, name

    def get(self, key, model=False):
        """Return the value of a single preference using a dotted path key"""
        section, name = self.parse_lookup(key)
        if model:
            return self.get_db_pref(setion=section, name=name)
        try:
            return self.from_cache(section, name)
        except CachedValueNotFound:
            pass

        db_pref = self.get_db_pref(section=section, name=name)
        self.to_cache(db_pref)
        return db_pref.value

    def get_db_pref(self, section, name):
        try:
            pref = self.queryset.get(section=section, name=name)

        except self.model.DoesNotExist:
            pref_obj = self.pref_obj(section=section, name=name)
            pref = self.create_db_pref(
                section=section, name=name, value=pref_obj.default)

        return pref

    def update_db_pref(self, section, name, value):
        try:
            db_pref = self.queryset.get(section=section, name=name)
            db_pref.value = value
            db_pref.save()
        except self.model.DoesNotExist:
            return self.create_db_pref(section, name, value)

        return db_pref

    def create_db_pref(self, section, name, value):
        if self.instance:
            db_pref = self.model(
                section=section, name=name, instance=self.instance)
        else:
            db_pref = self.model(section=section, name=name)
        db_pref.value = value
        db_pref.save()

        return db_pref

    def all(self):
        """Return a dictionnary containing all preferences by section
        Loaded from cache or from db in case of cold cache
        """
        a = {}
        try:
            for preference in self.registry.preferences():
                a[preference.identifier()] = self.from_cache(
                    preference.section, preference.name)
        except CachedValueNotFound:
            return self.load_from_db()

        return a

    def load_from_db(self):
        """Return a dictionnary of preferences by section directly from DB"""
        a = {}
        db_prefs = {p.preference.identifier(): p for p in self.queryset}
        for preference in self.registry.preferences():
            try:
                db_pref = db_prefs[preference.identifier()]
            except KeyError:
                db_pref = self.create_db_pref(
                    section=preference.section, name=preference.name, value=preference.default)

            self.to_cache(db_pref)
            a[preference.identifier()] = self.from_cache(
                preference.section, preference.name)

        return a



from django.apps import AppConfig

from . import global_preferences_registry
from .registries import autodiscover


class DynamicPreferencesConfig(AppConfig):
    name = 'dynamic_preferences'
    verbose_name = "Dynamic Preferences"

    def ready(self):
        autodiscover()

        



import collections


def update(d, u):
    """
        Custom recursive update of dictionary
        from http://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
    """
    for k, v in u.iteritems():
        if isinstance(v, collections.Mapping):
            r = update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


"""
Preferences are regular Python objects that can be declared within any django app.
Once declared and registered, they can be edited by admins (for :py:class:`SitePreference` and :py:class:`GlobalPreference`)
and regular Users (for :py:class:`UserPreference`)

UserPreference, SitePreference and GlobalPreference are mapped to corresponding PreferenceModel,
which store the actual values.

"""
from __future__ import unicode_literals

from .settings import preferences_settings

class AbstractPreference(object):
    """
    A base class that handle common logic for preferences
    """

    #: The section under which the preference will be registered
    section = None

    #: The preference name
    name = ""

    #: A default value for the preference
    default = None

    #: The model corresponding to this preference type (:py:class:`SitePreference`, :py:class:`GlobalPreference` or :py:class:`UserPreference`)
    model = None

    def __init__(self, registry=None):
        self.registry = registry

    @property
    def model(self):
        return self.registry.preference_model

    def identifier(self):
        """
        Return the name and the section of the Preference joined with a separator, with the form `section<separator>name`
        """
        section = self.section or ''
        if not section:
            return self.name
        return preferences_settings.SECTION_KEY_SEPARATOR.join([section, self.name])



# Taken from django-rest-framework
# https://github.com/tomchristie/django-rest-framework
# Copyright (c) 2011-2015, Tom Christie All rights reserved.

SETTINGS_ATTR = 'DYNAMIC_PREFERENCES'
USER_SETTINGS = None

try:
    from django.conf import settings
except ImportError:
    pass
else:
    # Only pull Django settings if Django environment variable exists.
    if settings.configured:
        USER_SETTINGS = getattr(settings, SETTINGS_ATTR, None)


DEFAULTS = {
    # 'REGISTRY_MODULE': 'prefs',
    # 'BASE_PREFIX': 'base',
    # 'SECTIONS_PREFIX': 'sections',
    # 'PREFERENCES_PREFIX': 'preferences',
    # 'PERMISSIONS_PREFIX': 'permissions',
    'MANAGER_ATTRIBUTE': 'preferences',
    'SECTION_KEY_SEPARATOR': '__',
    'REGISTRY_MODULE': 'dynamic_preferences_registry',
}


class PreferenceSettings(object):
    """
    A settings object, that allows API settings to be accessed as properties.
    For example:

        from rest_framework.settings import api_settings
        print(api_settings.DEFAULT_RENDERER_CLASSES)

    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.
    """
    def __init__(self, user_settings=None, defaults=None):
        self.user_settings = user_settings or {}
        self.defaults = defaults or DEFAULTS

    def __getattr__(self, attr):
        if attr not in self.defaults.keys():
            raise AttributeError("Invalid preference setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Cache the result
        setattr(self, attr, val)
        return val


preferences_settings = PreferenceSettings(USER_SETTINGS, DEFAULTS)





class DynamicPreferencesException(Exception):
    detail_default = 'An exception occurred with django-dynamic-preferences'

    def __init__(self, detail=None):
        if detail is not None:
            self.detail = str(detail)
        else:
            self.detail = str(self.detail_default)

    def __str__(self):
        return self.detail


class NotFoundInRegistry(DynamicPreferencesException, KeyError):
    detail_default = 'Preference with this name/section not found in registry'

class DoesNotExist(DynamicPreferencesException):
    detail_default = 'Cannot retrieve preference value, ensure the preference is correctly registered and database is synced'

class CachedValueNotFound(DynamicPreferencesException):
    detail_default = 'Cached value not found'



"""
    Class defined here are used by dynamic_preferences.preferences to specify
    preferences types (Bool, int, etc.) and rules according validation

"""
import datetime
from django import forms
from django.utils.translation import ugettext_lazy as _
from django.forms import CharField, IntegerField, BooleanField, ChoiceField, DateTimeField, TypedChoiceField
from django.utils.functional import cached_property

from .preferences import AbstractPreference
from dynamic_preferences.serializers import *

class BasePreferenceType(AbstractPreference):

    # A form field that will be used to display and edit the preference
    # use a class, not an instance
    field_class = None

    # these default will merged with ones from field_attributes
    # then pass to class provided in field in order to instantiate the actual field

    _default_field_attributes = {
        "required": True,
    }

    # Override this attribute to change field behaviour
    field_attributes = {}

    # A serializer class (see dynamic_preferences.serializers)
    serializer = None

    #: a default value or a callable that return a value to be used as default
    default_value = None

    _field = None        

    def get_field_kwargs(self):
        field_kwargs = dict(self._default_field_attributes)

        try:
            field_kwargs['initial'] = self.initial
        except AttributeError:
            pass
        field_kwargs.update(self.field_attributes)
        return field_kwargs

    @cached_property
    def default(self):
        """
        :return: If default_value is a a callable, return the callable result, else, return default_value as is
        """
        if hasattr(self.default_value, '__call__'):
            return self.default_value()
        else:
            return self.default_value

    @property
    def initial(self):
        """
        :return: initial data for form field, from field_attribute['initial'], _default_field_attribute['initial'] or
         default
        """
        return self.field_attributes.get('initial', self._default_field_attributes.get('initial', self.default))

    @property
    def field(self):
        return self.setup_field()

    def setup_field(self):
        """
            Create an actual instance of self.field
            Override this method if needed
        """
        
        return self.field_class(**self.get_field_kwargs())


class BooleanPreference(BasePreferenceType):

    BOOL_CHOICES = ((True, _('yes')), (False, _('no')))

    _default_field_attributes = {
        "choices": BOOL_CHOICES,
        "widget": forms.RadioSelect,
        "coerce": lambda x: x == 'True',
    }
    default = False
    field_class = TypedChoiceField
    serializer = BooleanSerializer


class IntPreference(BasePreferenceType):

    field_class = IntegerField
    serializer = IntSerializer


class StringPreference(BasePreferenceType):

    field_class = CharField
    serializer = StringSerializer
    default = ""

class LongStringPreference(StringPreference):
    _default_field_attributes = {
        "widget": forms.Textarea,
    }

class ChoicePreference(BasePreferenceType):

    choices = ()
    field_class = ChoiceField
    serializer = StringSerializer

    def get_field_kwargs(self):
        field_kwargs = super(ChoicePreference, self).get_field_kwargs()

        field_kwargs['choices'] = self.choices or self.field_attribute['initial']
        return field_kwargs




from django.conf.urls import patterns, include, url
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from . import views
from . import global_preferences_registry
from .forms import GlobalPreferenceForm


urlpatterns = patterns('',

    url(r'^global/$',
        staff_member_required(views.PreferenceFormView.as_view(
            registry=global_preferences_registry, 
            form_class=GlobalPreferenceForm)),
        name="dynamic_preferences.global"),
    url(r'^global/(?P<section>\w+)$',
        staff_member_required(views.PreferenceFormView.as_view(
            registry=global_preferences_registry, 
            form_class=GlobalPreferenceForm)),
        name="dynamic_preferences.global.section"),

    url(r'^user/$',
        login_required(views.UserPreferenceFormView.as_view()),
        name="dynamic_preferences.user"),
    url(r'^user/(?P<section>\w+)$',
        login_required(views.UserPreferenceFormView.as_view()),
        name="dynamic_preferences.user.section"),
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='GlobalPreferenceModel',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False, verbose_name='ID', auto_created=True)),
                ('section', models.TextField(blank=True, default=None, null=True, max_length=255, db_index=True)),
                ('name', models.TextField(max_length=255, db_index=True)),
                ('raw_value', models.TextField(blank=True, null=True)),
            ],
            options={
                'verbose_name_plural': 'global preferences',
                'verbose_name': 'global preference',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='UserPreferenceModel',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False, verbose_name='ID', auto_created=True)),
                ('section', models.TextField(blank=True, default=None, null=True, max_length=255, db_index=True)),
                ('name', models.TextField(max_length=255, db_index=True)),
                ('raw_value', models.TextField(blank=True, null=True)),
                ('instance', models.ForeignKey(to=settings.AUTH_USER_MODEL, related_name='preferences')),
            ],
            options={
                'verbose_name_plural': 'user preferences',
                'abstract': False,
                'verbose_name': 'user preference',
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='userpreferencemodel',
            unique_together=set([('instance', 'section', 'name')]),
        ),
        migrations.AlterUniqueTogether(
            name='globalpreferencemodel',
            unique_together=set([('section', 'name')]),
        ),
    ]





from django.conf import settings
from django.db.models.fields import FieldDoesNotExist

from django.utils.importlib import import_module
# import the logging library
import logging

try:
    # use Python3 reload
    from imp import reload

except:

    # we are on Python2
    pass

# Get an instance of a logger
logger = logging.getLogger(__name__)


#: The package where autodiscover will try to find preferences to register


from .managers import PreferencesManager
from .settings import preferences_settings
from .exceptions import NotFoundInRegistry

class PreferenceModelsRegistry(dict):

    """Store relationships beetween preferences model and preferences registry"""

    def register(self, preference_model, preference_registry):
        self[preference_model] = preference_registry
        preference_registry.preference_model = preference_model

        self.attach_manager(preference_model, preference_registry)

    def attach_manager(self, model, registry):
        if not hasattr(model, 'instance'):
            return

        def instance_getter(self):
            return registry.manager(instance=self)

        getter = property(instance_getter)
        instance_class = model._meta.get_field('instance').rel.to
        setattr(instance_class, preferences_settings.MANAGER_ATTRIBUTE, getter)

    def get_by_preference(self, preference):
        return self[preference.__class__]

    def get_by_instance(self, instance):
        """Return a preference registry using a model instance"""
        # we iterate throught registered preference models in order to get the instance class
        # and check if instance is and instance of this class
        for model, registry in self.items():
            try:
                instance_class = model._meta.get_field('instance').rel.to
                if isinstance(instance, instance_class):
                    return registry
                    break
            except FieldDoesNotExist:  # global preferences
                pass
        return None

preference_models = PreferenceModelsRegistry()


class PreferenceRegistry(dict):

    """
    Registries are special dictionaries that are used by dynamic-preferences to register and access your preferences.
    dynamic-preferences has one registry per Preference type:

    - :py:const:`user_preferences`
    - :py:const:`site_preferences`
    - :py:const:`global_preferences`

    In order to register preferences automatically, you must call :py:func:`autodiscover` in your URLconf.

    """

    #: a name to identify the registry
    name = "preferences_registry"
    preference_model = None

    def register(self, preference_class):
        """
        Store the given preference class in the registry.

        :param preference_class: a :py:class:`prefs.Preference` subclass
        """
        preference = preference_class(registry=self)
        try:
            self[preference.section][preference.name] = preference

        except KeyError:
            self[preference.section] = {}
            self[preference.section][preference.name] = preference

        return preference_class

    def get(self, name, section=None):
        """
        Returns a previously registered preference

        :param section: The section name under which the preference is registered
        :type section: str.
        :param name: The name of the preference. You can use dotted notation 'section.name' if you want to avoid providing section param
        :type name: str.
        :return: a :py:class:`prefs.BasePreference` instance
        """
        # try dotted notation
        try:
            section, name = name.split(
                preferences_settings.SECTION_KEY_SEPARATOR)
            return self[section][name]

        except ValueError:
            pass

        # use standard params
        try:
            return self[section][name]

        except:
            raise NotFoundInRegistry("No such preference in {0} with section={1} and name={2}".format(
                self.__class__.__name__, section, name))

    def manager(self, **kwargs):
        """Return a preference manager that can be used to retrieve preference values"""
        return PreferencesManager(registry=self, model=self.preference_model, **kwargs)

    def sections(self):
        """
        :return: a list of apps with registered preferences
        :rtype: list
        """

        return self.keys()

    def preferences(self, section=None):
        """
        Return a list of all registered preferences
        or a list of preferences registered for a given section

        :param section: The section name under which the preference is registered
        :type section: str.
        :return: a list of :py:class:`prefs.BasePreference` instances
        """

        if section is None:
            return [self[section][name] for section in self for name in self[section]]
        else:
            return [self[section][name] for name in self[section]]


class PerInstancePreferenceRegistry(PreferenceRegistry):
    pass


def clear():
    """
    Remove all data from registries
    """
    from .dynamic_preferences_registry import global_preferences_registry
    global_preferences_registry.clear()
    for model, registry in preference_models.items():
        registry.clear()


def autodiscover(force_reload=False):
    """
    Populate the registry by iterating through every section declared in :py:const:`settings.INSTALLED_APPS`.

    :param force_reload: if set to `True`, the method will reimport previously imported modules, if any
    :type force_reload: bool.
    """
    print('Dynamic-preferences: autodiscovering preferences...')
    if force_reload:
        clear()

    for app in settings.INSTALLED_APPS:
        # try to import self.package inside current app
        package = '{0}.{1}'.format(
            app, preferences_settings.REGISTRY_MODULE)
        try:
            # print('Dynamic-preferences: importing {0}...'.format(package))
            module = import_module(package)

            if force_reload:
                # mainly used in tests
                reload(module)

        except ImportError as e:
            pass
            # print('Dynamic-preferences: cannnot import {0}, {1}'.format(package, e))



from six import string_types
from django import forms
from django.core.exceptions import ValidationError

from . import global_preferences_registry, user_preferences_registry
from .models import GlobalPreferenceModel, UserPreferenceModel
from .exceptions import NotFoundInRegistry


class AbstractSinglePreferenceForm(forms.ModelForm):

    class Meta:
        fields = ('section', 'name', 'raw_value')

    def __init__(self, *args, **kwargs):

        self.instance = kwargs.get('instance')
        initial = {}
        if self.instance:
            initial['raw_value'] = self.instance.value
            kwargs['initial'] = initial
        super(AbstractSinglePreferenceForm, self).__init__(*args, **kwargs)

        if self.instance.name:
            self.fields['raw_value'] = self.instance.preference.setup_field()


    def clean(self):
        cleaned_data = super(AbstractSinglePreferenceForm, self).clean()
        try:
            self.instance.name, self.instance.section = cleaned_data['name'], cleaned_data['section']
        except KeyError: # changelist form
            pass
        try:
            self.instance.preference
        except NotFoundInRegistry:
            raise ValidationError(NotFoundInRegistry.detail_default)
        return self.cleaned_data

    def save(self, *args, **kwargs):
        self.instance.raw_value = self.instance.preference.serializer.serialize(self.cleaned_data['raw_value'])
        return super(AbstractSinglePreferenceForm, self).save(*args, **kwargs)

class SinglePerInstancePreferenceForm(AbstractSinglePreferenceForm):

    class Meta:
        fields = ('instance',) + AbstractSinglePreferenceForm.Meta.fields

    def clean(self):
        cleaned_data = super(AbstractSinglePreferenceForm, self).clean()
        try:
            self.instance.name, self.instance.section = cleaned_data['name'], cleaned_data['section']
        except KeyError: # changelist form
            pass
        i = cleaned_data.get('instance')
        if i:
            self.instance.instance = i
            try:
                self.instance.preference
            except NotFoundInRegistry:
                raise ValidationError(NotFoundInRegistry.detail_default)
        return self.cleaned_data

class GlobalSinglePreferenceForm(AbstractSinglePreferenceForm):

    class Meta:
        model = GlobalPreferenceModel
        fields = AbstractSinglePreferenceForm.Meta.fields


class UserSinglePreferenceForm(SinglePerInstancePreferenceForm):

    class Meta:
        model = UserPreferenceModel
        fields = SinglePerInstancePreferenceForm.Meta.fields


def preference_form_builder(form_base_class, preferences=[], **kwargs):
    """
    Return a form class for updating preferences
    :param form_base_class: a Form class used as the base. Must have a ``registry` attribute
    :param preferences: a list of :py:class:
    :param section: a section where the form builder will load preferences
    """
    registry = form_base_class.registry
    preferences_obj = []
    if len(preferences) > 0:
        # Preferences have been selected explicitly
        for pref in preferences:
            if isinstance(pref, string_types):
                preferences_obj.append(registry.get(name=pref))
            elif type(pref) == tuple:
                preferences_obj.append(
                    registry.get(name=pref[0], section=pref[1]))
            else:
                raise NotImplementedError(
                    "The data you provide can't be converted to a Preference object")
    elif kwargs.get('section', None):
        # Try to use section param
        preferences_obj = registry.preferences(
            section=kwargs.get('section', None))

    else:
        # display all preferences in the form
        preferences_obj = registry.preferences()

    fields = {}
    instances = []
    model_kwargs = kwargs.get('model', {})
    manager = registry.manager(**model_kwargs)

    for preference in preferences_obj:
        f = preference.field
        instance = manager.get_db_pref(
            section=preference.section, name=preference.name)
        f.initial = instance.value
        fields[preference.identifier()] = f
        instances.append(instance)

    form_class = type(
        'Custom' + form_base_class.__name__, (form_base_class,), {})
    form_class.base_fields = fields
    form_class.preferences = preferences_obj
    form_class.instances = instances
    return form_class


def global_preference_form_builder(preferences=[], **kwargs):
    """
    A shortcut :py:func:`preference_form_builder(GlobalPreferenceForm, preferences, **kwargs)`
    """
    return preference_form_builder(GlobalPreferenceForm, preferences, **kwargs)


def user_preference_form_builder(instance, preferences=[], **kwargs):
    """
    A shortcut :py:func:`preference_form_builder(UserPreferenceForm, preferences, **kwargs)`
    :param user: a :py:class:`django.contrib.auth.models.User` instance
    """
    return preference_form_builder(UserPreferenceForm, preferences, model={'instance': instance}, **kwargs)


class PreferenceForm(forms.Form):

    registry = None

    def update_preferences(self, **kwargs):
        for instance in self.instances:
            instance.value = self.cleaned_data[
                instance.preference.identifier()]
            instance.save()


class GlobalPreferenceForm(PreferenceForm):

    registry = global_preferences_registry


class UserPreferenceForm(PreferenceForm):

    registry = user_preferences_registry



"""
Preference models, queryset and managers that handle the logic for persisting preferences.
"""

from django.db import models
from django.db.models.query import QuerySet
from django.conf import settings
from django.utils.functional import cached_property

from dynamic_preferences import user_preferences_registry, global_preferences_registry
from dynamic_preferences.registries import preference_models
from .utils import update


class BasePreferenceModel(models.Model):

    """
    A base model with common logic for all preferences models.
    """

    #: The section under which the preference is declared
    section = models.TextField(
        max_length=255, db_index=True, blank=True, null=True, default=None)

    #: a name for the preference
    name = models.TextField(max_length=255, db_index=True)

    #: a value, serialized to a string. This field should not be accessed directly, use :py:attr:`BasePreferenceModel.value` instead
    raw_value = models.TextField(null=True, blank=True)

    class Meta:
        abstract = True
        app_label = 'dynamic_preferences'

    @cached_property
    def preference(self):
        return self.registry.get(section=self.section, name=self.name)
        

    def set_value(self, value):
        """
            Save serialized self.value to self.raw_value
        """
        self.raw_value = self.preference.serializer.serialize(value)

    def get_value(self):
        """
            Return deserialized self.raw_value
        """
        return self.preference.serializer.deserialize(self.raw_value)

    value = property(get_value, set_value)

    def save(self, **kwargs):

        if self.pk is None and not self.raw_value:

                self.value = self.preference.default
        super(BasePreferenceModel, self).save(**kwargs)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '{0} - {1}/{2}'.format(self.__class__.__name__, self.section, self.name)


class GlobalPreferenceModel(BasePreferenceModel):

    registry = global_preferences_registry

    class Meta:
        unique_together = ('section', 'name')
        app_label = 'dynamic_preferences'

        verbose_name = "global preference"
        verbose_name_plural = "global preferences"


class PerInstancePreferenceModel(BasePreferenceModel):

    """For preferences that are tied to a specific model instance"""
    #: the instance which is concerned by the preference
    #: use a ForeignKey pointing to the model of your choice
    instance = None

    class Meta(BasePreferenceModel.Meta):
        unique_together = ('instance', 'section', 'name')
        abstract = True

    @classmethod
    def get_instance_model(cls):
        return cls._meta.get_field('instance').rel.to

    @property
    def registry(self):
        return preference_models.get_by_instance(self.instance)


class UserPreferenceModel(PerInstancePreferenceModel):

    instance = models.ForeignKey(settings.AUTH_USER_MODEL)

    class Meta(PerInstancePreferenceModel.Meta):
        app_label = 'dynamic_preferences'
        verbose_name = "user preference"
        verbose_name_plural = "user preferences"


preference_models.register(UserPreferenceModel, user_preferences_registry)
preference_models.register(GlobalPreferenceModel, global_preferences_registry)

global_preferences_registry.preference_model = GlobalPreferenceModel

# Create default preferences for new instances

from django.db.models.signals import post_save


def create_default_per_instance_preferences(sender, created, instance, **kwargs):
    """Create default preferences for PerInstancePreferenceModel"""

    if created:
        try:
            registry = preference_models.get_by_instance(instance)
            registry.create_default_preferences(instance)
        except AttributeError:
            pass


def invalidate_cache(sender, created, instance, **kwargs):
    if not isinstance(instance, BasePreferenceModel):
        return
    registry = preference_models.get_by_preference(instance)
    linked_instance = getattr(instance, 'instance', None)
    kwargs = {}
    if linked_instance:
        kwargs['instance'] = linked_instance

    manager = registry.manager(**kwargs)
    manager.to_cache(instance)

post_save.connect(create_default_per_instance_preferences)
post_save.connect(invalidate_cache)



from .registries import PreferenceRegistry, PerInstancePreferenceRegistry


class GlobalPreferenceRegistry(PreferenceRegistry):
    def populate(self, **kwargs):
        return self.models(**kwargs)

class UserPreferenceRegistry(PerInstancePreferenceRegistry):
    pass


user_preferences_registry = UserPreferenceRegistry()
global_preferences_registry = GlobalPreferenceRegistry()
